/// A set of cache records.
public struct RecordSet: Hashable {
  public private(set) var storage: [CacheKey: Record] = [:]

  public init<S: Sequence>(records: S) where S.Iterator.Element == Record {
    insert(contentsOf: records)
  }

  public mutating func insert(_ record: Record) {
    storage[record.key] = record
  }
  
  public mutating func removeRecord(for key: CacheKey) {
    storage.removeValue(forKey: key)
  }

  public mutating func removeRecords(matching pattern: CacheKey) {
    for (key, _) in storage {
      if key.range(of: pattern, options: .caseInsensitive) != nil {
        storage.removeValue(forKey: key)
      }
    }
  }

  public mutating func clear() {
    storage.removeAll()
  }

  public mutating func insert<S: Sequence>(contentsOf records: S) where S.Iterator.Element == Record {
    for record in records {
      insert(record)
    }
  }

  public subscript(key: CacheKey) -> Record? {
    return storage[key]
  }

  public var isEmpty: Bool {
    return storage.isEmpty
  }

  public var keys: Set<CacheKey> {
    return Set(storage.keys)
  }

  @discardableResult public mutating func merge(records: RecordSet) -> Set<CacheKey> {
    var changedKeys: Set<CacheKey> = Set()

    for (_, record) in records.storage {
      changedKeys.formUnion(merge(record: record))
    }

    return changedKeys
  }

  @discardableResult public mutating func merge(record: Record) -> Set<CacheKey> {
    if var oldRecord = storage.removeValue(forKey: record.key) {
      var changedKeys: Set<CacheKey> = Set()

      for (key, value) in record.fields {
        if let oldValue = oldRecord.fields[key], oldValue == value {
          continue
        }
        oldRecord[key] = value
        changedKeys.insert([record.key, key].joined(separator: "."))
      }
      storage[record.key] = oldRecord
      return changedKeys
    } else {
      storage[record.key] = record
      return Set(record.fields.keys.map { [record.key, $0].joined(separator: ".") })
    }
  }
}

extension RecordSet: ExpressibleByDictionaryLiteral {
  public init(dictionaryLiteral elements: (CacheKey, Record.Fields)...) {
    self.init(records: elements.map { Record(key: $0.0, $0.1) })
  }
}

extension RecordSet: CustomStringConvertible {
  public var description: String {
    return String(describing: Array(storage.values))
  }
}

extension RecordSet: CustomDebugStringConvertible {
  public var debugDescription: String {
    return description
  }
}

extension RecordSet: CustomPlaygroundDisplayConvertible {
  public var playgroundDescription: Any {
    return description
  }
}
